#!/usr/bin/perl -w

use strict;
use warnings;
use threads;
use FindBin;
use lib "$FindBin::Bin/lib";
use Getopt::Long;
use File::Basename qw<basename dirname>;
use File::Spec;
use PTyper;
#use vars qw(%options);
my $usage = <<USAGE; 

=NAME

Pneumo-Typer.pl

=DESCRIPTION

    Run this command to enble users to obtain predict serotype and determine sequence type (ST/cgST) for Streptococcus pneumoniae, depending on obtaining the potential cps genes by a local blastn analysis using multiple threads. 

=USAGE

Pneumo-Typer.pl -dir genbank_file_directory [options]

FOR EXAMPLE: 

perl /Users/zilinyang/Desktop/桌面材料/李向阳-论文投稿汇总/Serotype_ST_manuscript/pneumo-typer/pneumo-typer/pneumo-typer.pl -d /Users/zilinyang/Desktop/桌面材料/李向阳-论文投稿汇总/Serotype_ST_manuscript/Pneumo-Typer_v1.03/final_93_gbk -t 10 -s T  -p T -o /Users/zilinyang/Desktop/桌面材料/李向阳-论文投稿汇总/Serotype_ST_manuscript/Pneumo-Typer_v1.03/final_93_gbk.OUT

perl /home/xiangyang/Mazhongrui/Streptococcus_pneumoniae/Pneumo-Typer_v1.03/Pneumo-Typer.pl -d /home/xiangyang/Mazhongrui/Streptococcus_pneumoniae/FIG_data -t 190 -s T -c T -m T -o /home/xiangyang/Mazhongrui/Streptococcus_pneumoniae/FIG_data_OUT

nohup perl /home/xiangyang/Mazhongrui/Streptococcus_pneumoniae/Pneumo-Typer_v1.03/Pneumo-Typer.pl -d /home/xiangyang/Mazhongrui/Streptococcus_pneumoniae/prokka_workplace/prokka_gbk_folder_plus72 -t 190 -b T -jsp T -jse T

#######################################################################################################################################
=ARGUMENTS
=======================
    REQUIRED ARGUMENTS:
    ~~~~~~~~~~~~~~~~~~~
       -d, --genbank_file_directory
           A directory containing annotated genomes as Genbank format file. To avoid a mistake, genome names cannot use special character,
           such as space, equal. For large number of genomes, users are recommended to download using Aspera, a high-speed file transfer
           tool (https://downloads.asperasoft.com/).                           
    OPTIONAL ARGUMENTS:
    ~~~~~~~~~~~~~~~~~~~
       -o, --output_directory
           An output directory holding all the generated files by Pneumo-Typer.pl. if this option is not set, interested_gene_generation.pl will create a directory named "interested_gene_workplace" in the bin directory from where AMRG_Anno.pl was invoked.
       -m, --multiple_threads
           Set thread number (Default: 1)
       -b, --start_at_blast 
           Jump to a local blastn analysis, and Skips sequencing extraction (Default: F).  
       -hgc, --homologous_gene_cutoff
           Set E-value, Identify, Coverage (Query and Subject), Match_length (alignment length) cutoff in Blastn analysis (default: E-value=1e-5, Identify=70, Coverage=90, Match_length=100).
       -s, --split_file
           Split the query file into serveral sub-query files for parallal works, and the spilt number equie to the used threads number.
       -jsp, --jump_split
           After splitting file, users can jump the process of spiltting file to perform blastn analysis again.
       -jse, --jump_serotype
           After blastn analysis, users can jump the process of doing blastn file to perform Serotype analysis again.
       -p, --prodigal_annotation
           Annotate all genomes using prodigal. 
       -m, --mlst
           Perform mlst analysis (Default: T). 
       -c, --cgmlst
           Perform cgmlst analysis. It need >10 mins for one genome (Default: F).
       -Ts, --test
           Run pneumo-typer using Test_data as input to check whether pneumo-typer is installed successfully (Default: F).
       -h, --help
           Show this message.

=AUTHOR

Dr. Xiangyang Li (E-mail: lixiangyang\@fudan.edu.cn), Kaili University; Bacterial Genome Data mining & Bioinformatic Analysis (www.microbialgenomic.cn/).

=COPYRIGHT

Copyright 2023, Xiangyang Li. All Rights Reserved.

USAGE


my %options = (
    'genbank_file_directory'                       => undef,   
    'workplace_directory'                          => undef,  
    'multiple_threads'                             => "3",
    'start_at_blast'                               => "F",
    'homologous_gene_cutoff'                       => "1e-5,70,90,100", #array to set blast parse cutoff: E-value, Identify, Coverage, Match_length
    'split_file'                                   => "T",
    'jump_split'                                   => "F",
    'jump_serotype'                                => "F",
    'prodigal_annotation'                          => "F",
    'mlst'                                         => "F",
    'cgmlst'                                       => "F",
    'test'                                         => "F",
    'help'                                         => undef

);
 

GetOptions(
    'd|genbank_file_directory=s'                     => \$options{genbank_file_directory},    
    'o|workplace_directory=s'                        => \$options{workplace_directory},
    't|multiple_threads=i'                           => \$options{multiple_threads}, 
    'b|start_at_blast=s'                             => \$options{start_at_blast},       
    'hgc|homologous_gene_cutoff=s'                   => \$options{homologous_gene_cutoff}, 
    's|split_file=s'                                 => \$options{split_file},
    'jsp|jump_split=s'                               => \$options{jump_split},
    'jse|jump_serotype=s'                            => \$options{jump_serotype},
    'p|prodigal_annotation=s'                        => \$options{prodigal_annotation},
    'm|mlst=s'                                       => \$options{mlst},
    'c|cgmlst=s'                                     => \$options{cgmlst},
    'Ts|test=s'                                      => \$options{test},
    'h|help'                                         => \$options{help}

);

if ( defined( $options{help} ) ) {
    print $usage;
    exit(0);
}


my $now_time = localtime;
print "\n$now_time: Pneumo-Typer.pl start...\n\n";

my $home_directory = $FindBin::Bin;   # obtaining the home directory where Pneumo-Typer.pl located
######################################################################################################
# Set the absolute path for four programs
my $blastn        = `which blastn`;  # "/usr/bin/blastp";
$blastn =~ s/\n//g; 

my $makeblastdb   = `which makeblastdb`; #"/usr/bin/makeblastdb";
$makeblastdb =~ s/\n//g; 
######################################################################################################


#check for Pneumo-Typer.pl workplace options
if ($options{test} eq "T") {
    $options{genbank_file_directory} = "$home_directory/test_data";
    &PTyper::check_tool($0, $options{genbank_file_directory});
    exit(1);  
}

#check for required options
if ( !( defined($options{genbank_file_directory}) ) ) {
    print $usage;
    exit(2);
}

my $path_genbank = File::Spec->rel2abs($options{genbank_file_directory});
my $db_file = "$home_directory/DATABASE/CPS_workplace/seq_file.fasta";
my $mappingList = "$home_directory/DATABASE/CPS_workplace/id_list";
my $thread_number = $options{multiple_threads};
my $prodigal_annotation = $options{prodigal_annotation};



my $workplace;
if ( defined( $options{workplace_directory} ) ) {
    $workplace = File::Spec->rel2abs($options{workplace_directory});
    mkdir $workplace;
}else {
    $workplace = "$home_directory/pneumo-pyper_workplace";
    $workplace =~ s/\/\//\//g;
    mkdir $workplace;
}

my $path_fa1 = $workplace."/Whole_fasta1"; #fasta sequences are extracted as contigs for each genome
mkdir $path_fa1;

my $path_fa2 = $workplace."/Whole_fasta2"; #fasta sequences are concatenated for each genome
mkdir $path_fa2;

my $path_gene = $workplace."/Whole_gene";
mkdir $path_gene;

my $directory_TFT = $workplace."/Whole_TFT"; 
mkdir $directory_TFT;


my $new_db_file = $workplace."/".basename($db_file);
system ("cp $db_file $new_db_file");
system ("$makeblastdb -in $new_db_file -dbtype nucl > $workplace/temp.txt");


if ($options{start_at_blast} eq "F"){
    print "STEP-1: Dealing with genomes <extract genome sequence, gene seqeunces and gene feature table (TFT); annnotate genome which has no annotation infotmation using prodigal>\n";

    &PTyper::batch_genomenucleotide_extract ($path_genbank, $path_fa1, $path_fa2, $thread_number);

    &PTyper::batch_genenucleotide_TFT_extract ($path_genbank, $directory_TFT, $path_gene, $thread_number, $prodigal_annotation, $path_fa1) if $prodigal_annotation eq "F";

    &PTyper::prodigal_bacth($path_fa1, $path_gene, $directory_TFT, $thread_number, $prodigal_annotation) if $prodigal_annotation eq "T";
}

print "\nSTEP-2: Determining the Sequence Type (ST)\n" if ( ($options{mlst} eq "T") or ($options{cgmlst} eq "T") );

if ($options{mlst} eq "T") {
    print "  \nSTEP-2.1: MLST analysis\n";
    system ("perl $home_directory/ST_tool/ST_profile.pl $path_genbank $path_gene $home_directory/ST_tool/database/mlst $thread_number $workplace/ST_workplace");
}

if ($options{cgmlst} eq "T") {
    print "  \nSTEP-2.2: cgMLST analysis\n";
    system ("perl $home_directory/ST_tool/ST_profile_cg.pl $path_genbank $path_fa2 $home_directory/ST_tool/database/cgmlst $thread_number $workplace/ST_workplace");
}

 
print "\nSTEP-3: Predicting serotype\n";
my ($e_value, $identify, $coverage, $match_length) = split /,/, $options{homologous_gene_cutoff};
chdir $path_gene;

system ("find . -name '*.fasta' | xargs cat > $workplace/all.all.fasta") if $options{jump_split} eq "F";  # For large number of files, directly using cat may causing error

my $total_gene = "$workplace/all.all.fasta";
my $best_output = "$workplace/BEST.BLASTOUT";

if ($options{split_file} eq "T"){

    my $split_protein_path = "$workplace/split_diretory";
    mkdir $split_protein_path;
    &PTyper::split_to_subfiles ($total_gene, $thread_number, $split_protein_path) if $options{jump_split} eq "F";

    opendir SPLIT_PROTEIN_PATH, $split_protein_path or die "could not open $split_protein_path";
    my @SPLIT_PROTEIN_PATH = readdir SPLIT_PROTEIN_PATH; 
    @SPLIT_PROTEIN_PATH =grep ($_!~/^\./ ,@SPLIT_PROTEIN_PATH);  #delete hidden file . ..
    @SPLIT_PROTEIN_PATH =grep ($_=~/\d+$/ ,@SPLIT_PROTEIN_PATH);  #delete hidden file . ..
    closedir SPLIT_PROTEIN_PATH;
    my (@input, @outfile);
    foreach my $splitprotein(@SPLIT_PROTEIN_PATH){ 
        my $input="$split_protein_path/$splitprotein"; 
        push (@input,$input);  
        my $output="$split_protein_path/$splitprotein.blastout";
        push (@outfile,$output);
    }

    my $file_number = scalar @input;
    if ($file_number <= $options{multiple_threads}) {
        $thread_number = $file_number;
    }

    my @sub_parameters = (\@input, $new_db_file, \@outfile, $e_value, $blastn);
    &PTyper::bacth_blast_best($file_number, $thread_number, "blastn_best", \@sub_parameters) if $options{jump_serotype} eq "F";
    chdir $split_protein_path;
    system ("find . -name '*.blastout' | xargs cat > $best_output");  # For large number of files, directly using cat may causing error
    #system ("rm -rf $split_protein_path"); #delete split_diretory used for split blastn analysis
}else {

    &PTyper::blastn_best ($total_gene, $new_db_file, $best_output, $e_value, $blastn) if $options{jump_serotype} eq "F";  # A vs A blast

}

#system("rm -f $total_gene");

my $all_vs_all_cluster = PTyper::blast_filter ($best_output, $e_value, $identify, $coverage, $match_length);


############################################################
############################################################
my $map_table = dirname ($new_db_file)."/CPS_Mapping_table.txt";
&PTyper::SARG_map_transformation($new_db_file, $mappingList, $map_table);

my $cluster_mapping;
$cluster_mapping = "$workplace/CPS_cluster_mapping.result";
&PTyper::mapping ($map_table, $all_vs_all_cluster, $cluster_mapping);

############################################################
############################################################

&PTyper::result_statistics($workplace, $directory_TFT, $cluster_mapping);

system ("perl $home_directory/script/serotype_correct.pl $home_directory/DATABASE/serotype_matrix.txt $workplace/result_statistics/Statistics_OUT/CPS_LocusTag_statistics $workplace/Serotype.out");

print "\nSTEP-4: Output sequence type and serotype results\n";

system ("perl $home_directory/script/ser_join_ST.pl $workplace/Serotype.out $workplace/ST_out.txt $workplace/cgST_out.txt $workplace/Serotype_ST.out");

system ("rm -rf $new_db_file*"); #delete all serotype database files


### Visualizing the result
##########################

print "\nSTEP-5: Heatmaping the cps gene distribution in genomes\n";
my $cmd_1 ="perl $home_directory/script/heatmap.pl -dir $workplace/result_statistics/tbl_heatmap_antibiotic -left 20 -scale 4 -label T -dis 9 -w 4 -l 0 -right 50 -cf $workplace/result_statistics/Statistics_OUT/classification_CPS -e $workplace/Serotype_ST.out -o $workplace";
system ("$cmd_1"); 


my $cmd_2 ="perl $home_directory/script/heatmap.pl -dir $workplace/result_statistics/tbl_heatmap_gene -left 20 -scale 4 -label T -dis 9 -w 4 -l 0 -right 50 -cf $workplace/result_statistics/Statistics_OUT/classification_gene -e $workplace/Serotype_ST.out -o $workplace";
system ("$cmd_2"); 


print "\nSTEP-6: Visualizing the cps gene cluster in each genome\n";

my $gcluster_workplace = "$workplace/Gcluster_workplace";
mkdir $gcluster_workplace;
my $subtft = "$gcluster_workplace/Sub_TFT";
mkdir $subtft;
#obtain interested gene and sub TFT files, in which gene distance from (20 kb) cps genes were removed; 
#only genome having the number of cps gene > 1 is shown 
#print "    Cmd: perl $home_directory/script/interested_gene.pl $home_directory/DATABASE/gene.txt $workplace/Serotype.out $workplace/result_statistics/tbl_part $gcluster_workplace/interested_gene.txt $subtft 20000 1\n\n";
system ("perl $home_directory/script/interested_gene.pl $home_directory/DATABASE/gene.txt $workplace/Serotype.out $workplace/result_statistics/tbl_part $gcluster_workplace/interested_gene.txt $subtft 20000 1");

#obtain homologs cluster
my $homologs_cluster = "$gcluster_workplace/blast_homologs_cluster";
mkdir $homologs_cluster;
system ("perl $home_directory/script/mcr.pl $workplace/CPS_cluster_mapping.result $homologs_cluster/all_orthomcl.out");

#print "    Cmd: perl $home_directory/script/Gcluster.pl -dir $path_genbank -gene $gcluster_workplace/interested_gene.txt -m $thread_number -map T -o $gcluster_workplace -SVG T -PNG F -n 40\n";
system ("perl $home_directory/script/Gcluster.pl -dir $path_genbank -gene $gcluster_workplace/interested_gene.txt -m $thread_number -map T -o $gcluster_workplace -SVG T -PNG F -n 40"); 


my $finish_time = localtime;
print "\n$finish_time: done!\n\n";


